/* mge-utalk.c 

   New style driver for MGE UPS with UTALK protocol

   See doc/mge-utalk.txt for design information.

   Version 0.50.0b, 2002-05-07

   Copyright (C) 2002 Hans Ekkehard Plesser <hans.plesser@itf.nlh.no>

   This driver is a collaborative effort by

     Hans Ekkehard Plesser <hans.plesser@itf.nlh.no>
     Dave Abbott <d.abbott@dcs.shef.ac.uk>
     Arnaud Quette <arnaud.quette@mgeups.com>
     Nicholas Reilly <nreilly@magma.ca>
     Marek Kralewski <marek@mercy49.de>

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
*/

#include <time.h>
#include <ctype.h>
#include "main.h"
#include "mge-utalk.h"

/* --------------------------------------------------------------- */
/*                  Define "technical" constants                   */
/* --------------------------------------------------------------- */

#define DRIVER_NAME    "MGE UPS SYSTEMS/U-Talk driver"
#define DRIVER_VERSION "0.51.0"

#define MGE_CHAR_DELAY 0     /* delay after sending each char to UPS,
                                in MICROSECONDS                    */
#define MGE_REPLY_DELAY 1000 /* delay after command, before reading
                                  UPS reply, in MICROSECONDS       */
#define MGE_CONNECT_DELAY 500000 /* delay after enable_ups_comm    */

#define MGE_COMMAND_ENDCHAR "\r\n"   /* some UPS need \r and \n */
#define MGE_REPLY_ENDCHAR   '\r'
#define MGE_REPLY_IGNCHAR   "\r\n"

#define MAXTRIES    10 /* max number of connect tries              */
#define BUFFLEN    256 

extern int  sddelay;         /* from main.c --- ignored for now  */

/* --------------------------------------------------------------- */
/*             Structure with information about UPS                */
/* --------------------------------------------------------------- */

static struct {
  int MultTab; 
  int LowBatt;  /* percent */
  int OnDelay;  /* minutes */
  int OffDelay; /* seconds */
} mge_ups = { 0, DEFAULT_LOWBATT, DEFAULT_ONDELAY, DEFAULT_OFFDELAY };


/* --------------------------------------------------------------- */
/*             Declaration of internal functions                   */
/* --------------------------------------------------------------- */

static void instcmd (int auxcmd, int dlen, char *data);
static void setvar (int auxcmd, int dlen, char *data);
static void enable_ups_comm();
static void extract_info(const char *buf, const mge_info_item *mge, 
			 char *infostr, int infolen);
static char *info_variable_name(int type);
static char *info_variable_cmd(int type);
static bool info_variable_ok(int type);
static int  get_ups_status();
static int mge_command(char *reply, int replylen, const char *fmt, ...);
static void format_model_name(char *model);

/* --------------------------------------------------------------- */
/*                    UPS Driver Functions                         */
/* --------------------------------------------------------------- */

void upsdrv_banner(void)
{
  printf("Network UPS Tools - %s %s (%s)\n", 
	 DRIVER_NAME, DRIVER_VERSION, UPS_VERSION);
}

/* --------------------------------------------------------------- */

void upsdrv_makevartable(void)
{
  char temp[BUFFLEN];

  /* add command line/conf variables */
  /* help looks nicer if all comments have same length */
  snprintf(temp, sizeof(temp), 
	   "Low battery level, in %%              (default = %3d)", 
	   DEFAULT_LOWBATT);
  addvar (VAR_VALUE, "LowBatt", temp);

  snprintf(temp, sizeof(temp), 
	   "Set delay before startup, in minutes (default = %3d)",
	   DEFAULT_ONDELAY);
  addvar (VAR_VALUE, "OnDelay", temp);

  snprintf(temp, sizeof(temp), 
	   "Delay before shutdown, in seconds    (default = %3d)",
	  DEFAULT_OFFDELAY);
  addvar (VAR_VALUE, "OffDelay", temp);
}

/* --------------------------------------------------------------- */

void upsdrv_initups(void)
{
  char buf[BUFFLEN];

  open_serial(device_path, B2400);
  enable_ups_comm();

  /* read command line/conf variables */
  if (getval ("LowBatt"))
    mge_ups.LowBatt = atoi (getval ("LowBatt"));

  if (getval ("OnDelay"))
    mge_ups.OnDelay = atoi (getval ("OnDelay"));

  if (getval ("OffDelay"))
    mge_ups.OffDelay = atoi (getval ("OffDelay"));

  /* set values given on the command line */
  mge_command(buf, sizeof(buf), "Bl %d",  mge_ups.LowBatt);
  if(strcmp(buf, "OK")) {
    upslogx(LOG_INFO, "UPS response to %d%% Low Batt Level was %s", 
	    mge_ups.LowBatt, buf);
  }

  mge_command(buf, sizeof(buf), "Sm %d",  mge_ups.OnDelay);
  if(strcmp(buf, "OK")) {
    upslogx(LOG_INFO, "UPS response to %d min ON delay was %s", 
	    mge_ups.OnDelay, buf);
  }

  mge_command(buf, sizeof(buf), "Sn %d",  mge_ups.OffDelay);
  if(strcmp(buf, "OK")) {
    upslogx(LOG_INFO, "UPS response to %d min OFF delay was %s", 
	    mge_ups.OffDelay, buf);
  }

}

/* --------------------------------------------------------------- */

int upsdrv_infomax(void)
{
  int k;

  /* count elements in mge_info_item */
  k = 0;
  while( mge_info[k].type != INFO_UNUSED )
    k++;

  /* add MEMBERS, MFR, UPSIDENT, MODEL, FIRMWARE, STATUS */
  k += 6;

  /* add a little slack, in case I miscounted  */
  k += 6;

  return k;

}

/* --------------------------------------------------------------- */

void upsdrv_initinfo(void)
{

  char buf[BUFFLEN];
  char *model = NULL;
  char *firmware = NULL;
  char *p;
  int  table;
  int  tries;
  int  status_ok;
  int  bytes_rcvd;
  mge_info_item *item;

  /* manufacturer -------------------------------------------- */
  addinfo (INFO_MFR, "MGE UPS SYSTEMS",  FLAG_STRING, 15);

  /* model, firmware, status, multiplier table --------------- */
  addinfo(INFO_STATUS,   "", 0, 0);  /* required, so add here  */

  /* loop until we have at status */
  tries = 0;
  do {
    printf(".");

    /* get model information: <Family> <Model> <Firmware> */
    bytes_rcvd = mge_command(buf, sizeof(buf), "Si 1");
    if(bytes_rcvd > 0 && buf[0] != '?') {

/*        addinfo(INFO_UPSIDENT, "", 0, 0); */
      addinfo(INFO_UPSIDENT, "", FLAG_STRING, 15);
      setinfo(INFO_UPSIDENT, "%s", buf); /* raw id */

      model = buf;
      p = strrchr(buf, ' ');
      if ( p != NULL ) {
	*p = '\0';
	firmware = p+1;
      }

      if( firmware && strlen(firmware) < 1 )
	firmware = NULL;   /* no firmware information */
    }

    if ( model ) {
      format_model_name(model);
      addinfo(INFO_MODEL,    "", 0, 0);
      setinfo(INFO_MODEL,   "%s", model);
    }

    if ( firmware ) {
      addinfo(INFO_FIRMREV,  "", 0, 0);
      setinfo(INFO_FIRMREV, "%s", firmware);
    }

    /* multiplier table */
    /* <protocol level> <multiplier table> */ 
    bytes_rcvd = mge_command(buf, sizeof(buf), "Ai");
    if (bytes_rcvd > 0 && buf[0] != '?') {
      p = strchr(buf, ' ');
      if ( p != NULL ) {
	table = atoi(p + 1);
    	if ( 0 < table && table < 4 )
	  mge_ups.MultTab = table;
      }
    }
    
    /* status --- try only system status, to get the really important
       information (OL, OB, LB); all else is added later by updateinfo */
    status_ok = get_ups_status();

  } while ( !status_ok && tries++ < MAXTRIES );
  
  if ( tries == MAXTRIES && !status_ok )
    fatalx("Could not get status from UPS.");
  if ( mge_ups.MultTab == 0 )
    upslogx(LOG_WARNING, 
	    "Could not get multiplier table: using raw readings.");


  /* all other variables ------------------------------------ */
  for ( item = mge_info ; item->type != INFO_UNUSED ; item++ ) {

    char infostr[VALSIZE];
    int  chars_rcvd;

    /* send request, read answer */
    chars_rcvd = mge_command(buf, sizeof(buf), item->cmd);

    if ( chars_rcvd < 1 || buf[0] == '?' ) {

      item->ok = FALSE;
      upsdebugx(LOG_INFO, "initinfo: %s unavailable", 
	      info_variable_name(item->type));
      
    } else {

      item->ok = TRUE;
      extract_info(buf, item, infostr, sizeof(infostr));
      addinfo(item->type, infostr, item->flags, item->length);

      upsdebugx(1, "initinfo: %s == >%s<",
	       info_variable_name(item->type), infostr);

    }
     
  } /* for item */

  /* commands ----------------------------------------------- */
  addinfo (INFO_INSTCMD, "", 0, CMD_BTEST1);
  addinfo (INFO_INSTCMD, "", 0, CMD_FPTEST);
  addinfo (INFO_INSTCMD, "", 0, CMD_SHUTDOWN);
  addinfo (INFO_INSTCMD, "", 0, CMD_OFF);
  addinfo (INFO_INSTCMD, "", 0, CMD_ON);
  addinfo (INFO_INSTCMD, "", 0, CMD_BYPASS);

  /* install handlers */
  upsh.instcmd = instcmd;
  upsh.setvar = setvar;
  
  /* call writeinfo since we have already collected a lot */
  writeinfo();

}

/* --------------------------------------------------------------- */

void upsdrv_updateinfo(void)
{
  char buf[BUFFLEN];
  int status_ok;
  mge_info_item *item;

  /* make sure that communication is enabled */
  enable_ups_comm();
   
  /* update status */
  status_ok = get_ups_status();  /* only sys status is critical */
  if ( !status_ok )
    upslogx(LOG_NOTICE, "updateinfo: Cannot update system status");

  /* update all other ok variables */
  for ( item = mge_info ; item->type != INFO_UNUSED ; item++ ) 
    if ( item->ok ) {

      int bytes_rcvd;

      /* send request, read answer */
      bytes_rcvd = mge_command(buf, sizeof(buf), item->cmd);

      if ( bytes_rcvd > 0 && buf[0] != '?' )  {

	char infostr[VALSIZE];
	extract_info(buf, item, infostr, sizeof(infostr));
	setinfo(item->type, infostr);
	upsdebugx(2, "updateinfo: %s == >%s<",
		  info_variable_name(item->type), infostr);

      } else {

	upslogx(LOG_NOTICE, "updateinfo: Cannot update %s", 
		info_variable_name(item->type));
      
      }

    } /* if item->ok */

  writeinfo();

}

/* --------------------------------------------------------------- */

void upsdrv_shutdown(void)
{
  char buf[BUFFLEN];
/*  static time_t lastcmd = 0; */

  /* enable automatic restart */
  mge_command(buf, sizeof(buf), "Sx 5");
  if(strcmp(buf, "OK")) {
    upslogx(LOG_INFO, "UPS response to Automatic Restart was %s", 
	    buf);
  }

  /* shutdown UPS */
  mge_command(buf, sizeof(buf), "Sx 0");
  if(strcmp(buf, "OK")) {
    upslogx(LOG_INFO, "UPS response to Shutdown was %s", 
	    buf);
  }
}

/* --------------------------------------------------------------- */

void upsdrv_help(void)
{
  printf("\n");
  printf("Option -d is ignored.\n");
  printf("Shutdown command must be confirmed within %d to %d seconds\n",
	 MIN_CONFIRM_TIME, MAX_CONFIRM_TIME);
  printf("to become effective.  The UPS is restarted when power returns.\n"); 
  printf("\n");
}


/* --------------------------------------------------------------- */
/*                      Internal Functions                         */
/* --------------------------------------------------------------- */

/* deal with truncated model names */
void format_model_name(char *model)
{
  if(!strncmp(model, "Evolutio", 8))
    sprintf(model, "Evolution %i", atoi(strchr(model, ' ')));
}

/* --------------------------------------------------------------- */

/* handler for commands to be sent to UPS */
void instcmd (int auxcmd, int dlen, char *data)
{
  char temp[BUFFLEN];

  switch (auxcmd) {

    case CMD_BTEST1: /*	start battery test */
      mge_command(temp, sizeof(temp), "Bx 1");
      if(strcmp(temp, "OK")) {
	upslogx(LOG_INFO, "UPS response to Start Battery Test was %s", temp);
      }
      /* TODO: read results */
      break;

    case CMD_FPTEST:	 /* start front panel test  */
      mge_command(temp, sizeof(temp), "Sx 129");
      if(strcmp(temp, "OK")) {
	upslogx(LOG_INFO, 
		"UPS response to Start Front Panel Test was %s", temp);
      }
      break;

  case CMD_SHUTDOWN:
      upsdrv_shutdown();
      break;

  case CMD_OFF: /* Power Off all plugs */
    /* TODO: control per plug */
      mge_command(temp, sizeof(temp), "Wy 65535");
      upslogx(LOG_INFO, "UPS response to Select All Plugs was %s", temp);

      if(!strcmp(temp, "OK")) {
	mge_command(temp, sizeof(temp), "Wx 0");
	upslogx(LOG_INFO, "UPS response to Power Off Selected Plugs was %s", temp);
      }
      break;
  case CMD_ON: /* Power On all plugs */
    /* TODO: control per plug */
      mge_command(temp, sizeof(temp), "Wy 65535");
      upslogx(LOG_INFO, "UPS response to Select All Plugs was %s", temp);

      if(!strcmp(temp, "OK")) {      
	mge_command(temp, sizeof(temp), "Wx 1");
	upslogx(LOG_INFO, "UPS response to Power On Selected Plugs was %s", temp);
      }
      break;

  case CMD_BYPASS: /* switch on/off Maintenance Bypass */
    /* read maintenance bypass status */
    if(mge_command(temp, sizeof(temp), "Ps") >0) {
	if (temp[0] == '1') {
	  mge_command(temp, sizeof(temp), "Px 2"); /* Disable Maintenance Bypass */
	  upslogx(LOG_INFO, "UPS response to Disable Maintenance Bypass was %s", temp);
	}
	else {
	  mge_command(temp, sizeof(temp), "Px 3"); /* Enable Maintenance Bypass */
	  upslogx(LOG_INFO, "UPS response to Enable Maintenance Bypass was %s", temp);
	}
    } 
    break;
	
    default:
      upslogx(LOG_INFO, "instcmd: unknown type 0x%04x", auxcmd);
      break;
  }

}
/* --------------------------------------------------------------- */

/* handler for settable variables in UPS*/
void setvar (int auxcmd, int dlen, char *data)
{
  char temp[BUFFLEN];
  char cmd[15];

  /* TODO : add some controls */

  if(info_variable_ok(auxcmd)) {
    /* format command */
    sprintf(cmd, "%s", info_variable_cmd(auxcmd));
    sprintf(strchr(cmd, '?'), "%s", data);
    
    /* Execute command */
    mge_command(temp, sizeof(temp), cmd);
    upslogx(LOG_INFO, "UPS response to Set %s to %s was %s", 
	    info_variable_name(auxcmd), data, temp);
  }
  else
    upslogx(LOG_ERR, "ERROR: Variable %s not supported by UPS", 
	    info_variable_name(auxcmd));
}
/* --------------------------------------------------------------- */

/* enable communication with UPS */
static void enable_ups_comm()

{
  mge_command(NULL, 0, "Z");   /* send Z twice --- speeds up re-connect */
  mge_command(NULL, 0, "Z");   
  mge_command(NULL, 0, "Ax 1");   
  usleep(MGE_CONNECT_DELAY);
  upsflushin(0, nut_debug_level, "?\r\n");  /* ignore ? as well " */
}

/* --------------------------------------------------------------- */

/* extract information from buffer
   in:   buf    : reply from UPS
         item   : INFO item queried
   out:  infostr: to be placed in INFO_ variable 
   NOTE: buf="?" must be handled before calling extract_info
         buf is changed inspite of const !!!!!
*/
static void extract_info(const char *buf, const mge_info_item *item, 
			 char *infostr, int infolen)
{
  /* initialize info string */
  infostr[0] = '\0';

  /* write into infostr with proper formatting */
  if ( strpbrk(item->fmt, "feEgG") ) {           /* float */
    snprintf(infostr, infolen, item->fmt,
	     multiplier[mge_ups.MultTab][item->unit] * atof(buf));
  } else if ( strpbrk(item->fmt, "dioxXuc") ) {  /* int   */
    snprintf(infostr, infolen, item->fmt,
	     (int) (multiplier[mge_ups.MultTab][item->unit] * atof(buf)));
  } else {
    snprintf(infostr, infolen, item->fmt, buf);
  }
}



/* --------------------------------------------------------------- */

/* get system status, at least: OB, OL, LB
   calls set_status appropriately
   tries MAXTRIES times
   returns non-nil if successful           

   NOTE: MGE counts bytes/chars the opposite way as C, 
         see mge-utalk.txt.  If status commands send two
         data items, these are separated by a space, so
	 the elements of the second item are in buf[16..9].
*/

static int get_ups_status()
{
  char buf[BUFFLEN]; 
  int rb_set= FALSE;  /* has RB flag been set ? */
  int over_set= FALSE;  /* has OVER flag been set ? */
  int tries = 0;
  int ok    = FALSE;
  int bytes_rcvd = 0;

  do {
    /* must clear status buffer before each round */
    status_init();

    /* system status */
    bytes_rcvd = mge_command(buf, sizeof(buf), "Ss");
    upsdebugx(1, "Syst Stat >%s<", buf);
    if ( bytes_rcvd > 0 && strlen(buf) > 7 ) {
      ok = TRUE;
      if (buf[6] == '1') {
	over_set = TRUE;
	status_set("OVER");
      }
      if (buf[5] == '1') {
	status_set("OB");
      } else {
	status_set("OL");
      }
      if (buf[4] == '1') {
	status_set("LB");
      }
      if (buf[3] == '1') {
	rb_set = TRUE;
	status_set("RB"); 
      }
      /* buf[2] not used */
      if (buf[1] == '1') {
	status_set("COMMFAULT"); /* self-invented */
      }
      if (buf[0] == '1') {
	status_set("ALARM");     /* self-invented */
      } 
    }  /* if strlen */

    /* battery status */
    mge_command(buf, sizeof(buf), "Bs");
    upsdebugx(1, "Batt Stat >%s<", buf);
    if ( strlen(buf) > 7 ) {
      if ( !rb_set && ( buf[7] == '1' || buf[3] == '1' ) ) {
	status_set("RB");
      }
      if (buf[1] == '1') {
	status_set("CHRG");
      } 
      if (buf[0] == '1') {
	status_set("DISCHRG");
      }
    } /* if strlen */      
      
    /* load status */
    mge_command(buf, sizeof(buf), "Ls");
    upsdebugx(1, "Load Stat >%s<", buf);
    if ( strlen(buf) > 7 ) {
      if (buf[4] == '1') {
	status_set("BOOST");
      } 
      if ( !over_set && ( buf[3] == '1' ) ) {
	status_set("OVER");
      }
      if (buf[2] == '1') {
	status_set("TRIM");
      }
    } /* if strlen */      
    if ( strlen(buf) > 15 ) {   /* second "byte", skip <SP> */
      if (buf[16] == '1') {
	status_set("FSD");
      } 
      if ( !(buf[9] == '1') ) {
	status_set("OFF");
      }
    } /* if strlen */  
  } while ( !ok && tries++ < MAXTRIES );

  status_commit();

  return ok;

}

/* --------------------------------------------------------------- */

/* return proper variable "ok" given INFO_ type */
   
static bool info_variable_ok(int type) 
{
  mge_info_item *item = mge_info ;

  while ( item->type != type )
    item++;

  return item->ok;
}

/* --------------------------------------------------------------- */

/* return proper variable "cmd" given INFO_ type */
   
static char *info_variable_cmd(int type) 
{
  mge_info_item *item = mge_info ;

  while ( item->type != type )
    item++;

  return item->cmd;
}

/* --------------------------------------------------------------- */

/* return proper variable name given INFO_ type */
   
static char *info_variable_name(int type) 
{
  struct netvars_t *var = netvars;

  while ( var->type != type )
    var++;

  return var->name;
}


/* --------------------------------------------------------------- */

/* send command to UPS and read reply if requested
   
   reply   :  buffer for reply, NULL if no reply expected
   replylen:  length of buffer reply
   fmt     :  format string, followed by optional data for command

   returns :  no of chars received, -1 if error
*/
static int mge_command(char *reply, int replylen, const char *fmt, ...)
{
  char *p;
  char command[BUFFLEN];
  int bytes_sent = 0;
  int bytes_rcvd = 0;
  va_list ap;

  /* build command string */
  va_start(ap, fmt);
  if (vsnprintf(command, sizeof(command), fmt, ap) >= sizeof(command))
    upsdebugx(4, "mge_command: command truncated");
  va_end(ap);

  /* flush received, unread data */
  tcflush(upsfd, TCIFLUSH);

  /* send command */
  for (p = command; *p; p++) {
    if ( isprint(*p) )
      upsdebugx(4, "mge_command: sending [%c]", *p);
    else
      upsdebugx(4, "mge_command: sending [%02X]", *p);

    if (write(upsfd, p, 1) != 1)
      return -1;
    bytes_sent++;
    usleep(MGE_CHAR_DELAY);
  }

  /* send terminating string */
  if (MGE_COMMAND_ENDCHAR) {
    for (p = MGE_COMMAND_ENDCHAR; *p; p++) {

      if ( isprint(*p) )
	upsdebugx(4, "mge_command: sending [%c]", *p);
      else
	upsdebugx(4, "mge_command: sending [%02X]", *p);

      if (write(upsfd, p, 1) != 1)
	return -1;
      bytes_sent++;
      usleep(MGE_CHAR_DELAY);
    }
  }

  if ( !reply )
    return bytes_rcvd;
  else
    usleep(MGE_REPLY_DELAY);
  
  bytes_rcvd = upsrecv(reply, replylen, 
		       MGE_REPLY_ENDCHAR, MGE_REPLY_IGNCHAR);

  return bytes_rcvd;

}
